#!/usr/bin/perl

# Route-B get script
# by SUNAGAWA Hiroyuki / Studio JamPAck
# ver.0.21 (2021/01/26)
#
# USAGE:
# ex: ./routeb.pl [($COMMAND_str)[, ($SERIAL_dev)]]
#   $COMMAND_str: get, start

# モジュール読み込み
use Time::Local;
use Device::SerialPort;
use Time::HiRes 'sleep';


# メモ
# エラーが発生しやすい場所
# l.453 SKSENDTO後の待ちで応答がない？
# l.160 ROPT後の待ち


### 定義

# デバッグモード
#$SW_DEBUG = 1;
$SW_DEBUG = 0;
$DEBUGLOG = "";

# 標準出力などをコマンドバッファリングとする
$| = 1;
# 最大待ち時間
$EXECTTL = 20; # sec
# SKPING待ち時間
$ROUTEB_PINGWAIT = 2;

# シリアルポートは引数1から指定する
# ただし指定しなかった場合はttyUSB0とする
if($ARGV[1] =~ /^\/dev\//) {
  $SERIAL_dev = $ARGV[1];
} else {
  $SERIAL_dev = "/dev/ttyUSB0";
}
# ポートオープン時の待ち時間
$SERIAL_quiet = 1.0; # sec
# 読み込みバッファサイズ
$SERIAL_bytes = 1024; # char

# Bルート認証ID
$ROUTEB_ID = "********************************";
# Bルート認証パスワード
$ROUTEB_PWD = "************";

# SKSCANタイムアウト（だいたい4～6秒でいけるはず）
$ROUTEB_SCANTTL = 6;
# SKSCAN待ち時間
$ROUTEB_SCANWAIT = 1;

# スマートメーターアドレス一時ファイル
$ROUTEB_TEMPDIR = "/run/routeb";
$ROUTEB_TEMPFILE = $ROUTEB_TEMPDIR . "/addr.txt";
$ROUTEB_ERRORLOG = $ROUTEB_TEMPDIR . "/error.txt";

# 送受信用変数
my $SERIAL_rx_str, $SERIAL_rx_bytes;
my $SERIAL_tx_str, $SERIAL_tx_bytes, $SERIAL_tx_bin;
my $LOOPSTART;



### 処理開始

# ポートオープン
&debuglog("# opening device $SERIAL_dev\n");
$objtty = new Device::SerialPort($SERIAL_dev, $SERIAL_quiet)
  or &errorexit("Cannot open device $SERIAL_dev $!", __LINE__);

# ポート設定
$objtty->baudrate(115200);
$objtty->databits(8);
#$dbjtty->parity("none"); # 何故か対応していないと言われたので。。。
$objtty->stopbits(1);
$objtty->handshake("none");
&debuglog("# opening serial port device $SERIAL_dev\n");
$objtty->write_settings()
  or &errorexit("Cannot setup device $SERIAL_dev $!", __LINE__);

# キャッシュファイルからRoute-Bアドレス(rb_addr)取得
&debuglog("# reading rb_addr cached file\n");
if(-e $ROUTEB_TEMPFILE) {
  open FILE_rb_addr, "<$ROUTEB_TEMPFILE"
    or &errorexit("Cannot open file $ROUTEB_TEMPFILE $!", __LINE__);
  $rb_addr = <FILE_rb_addr>
    or &errorexit("Cannot read from file $ROUTEB_TEMPFILE $!", __LINE__);
  close FILE_rb_addr;
  &debuglog("<$ROUTEB_TEMPFILE : rb_addr = $rb_addr\n");
} else {
  $rb_addr = "";
  &debuglog("# cannot open rb_addr cached file $ROUTEB_TEMPFILE\n");
}

# 溜まっている受信バッファをすべて流す（必要なのか不明だがおまじないで入れてる？）
&debuglog("# flushing buffer\n");
$LOOPSTART = time;
do {
  ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
    or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
  if($SERIAL_rx_bytes > 0) {
    &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
#  } else {
#    &debuglog(".");
  }
} until($SERIAL_rx_bytes == 0 || $LOOPSTART + $EXECTTL < time);
if($LOOPSTART + $EXECTTL < time) {
  &errorexit("timeout", __LINE__);
}

# 疎通確認(SKPING)
if($rb_addr ne '') {
  &debuglog("# sending ping\n");
  $SERIAL_tx_str = "SKPING 0 $rb_addr\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $LOOPSTART = time;
  # 疎通応答(EPONG)待ち
  do {
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
#    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_str =~ /EPONG/ || $LOOPSTART + $ROUTEB_PINGWAIT < time);
  # 応答がなかったらrb_addrを破棄して初期化
  if($LOOPSTART + $ROUTEB_PINGWAIT < time) {
    $rb_addr = '';
    &debuglog("# clear cached rb_addr\n");
  }

}


### 初期化

if($rb_addr eq '') {
  
  # 初期化(SKRESET)
  &debuglog("# reset\n");
  $SERIAL_tx_str = "SKRESET\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $LOOPSTART = time;
  # 応答(OK)待ち
  do {
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
#    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_str =~ /OK/ || $LOOPSTART + $EXECTTL < time);
  if($LOOPSTART + $EXECTTL < time) {
    &errorexit("timeout", __LINE__);
  }
  
  # パラメータ設定（製品に対して1回のみ設定）
  
  # プロダクト設定確認(ROPT)
  &debuglog("# getting product option\n");
  $SERIAL_tx_str = "ROPT\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $LOOPSTART = time;
  # 応答(OK)待ち
  do {
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or die "ERROR: cannot read device $SERIAL_dev $!";
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
#    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_str =~ /OK (\d\d)/ || $LOOPSTART + $EXECTTL < time);
  if($LOOPSTART + $EXECTTL < time) {
    &errorexit("timeout", __LINE__);
  }
  &debuglog("# ROPT $1\n");
  
  # プロダクト設定（WOPT 01 : ROPTが00だった場合のみ）
  if($1 eq '00') {
    &debuglog("# setting product option\n");
    $SERIAL_tx_str = "WOPT 01\r\n";
    &debuglog("> $SERIAL_tx_str");
    $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
      or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
    $LOOPSTART = time;
    # 応答(OK)待ち
    do {
      ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
        or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
      if($SERIAL_rx_bytes > 0) {
        &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
      } else {
#        &debuglog(".");
      }
    } until($SERIAL_rx_str =~ /OK/ || $LOOPSTART + $EXECTTL < time);
    if($LOOPSTART + $EXECTTL < time) {
      &errorexit("timeout", __LINE__);
    }
  }
  
  # パラメーター設定（毎回設定する内容）
  
  # RBID
  &debuglog("# setting RBID\n");
  $SERIAL_tx_str = "SKSETRBID $ROUTEB_ID\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $LOOPSTART = time;
  do {
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_str =~ /OK/ || $LOOPSTART + $EXECTTL < time);
  if($LOOPSTART + $EXECTTL < time) {
    &errorexit("timeout", __LINE__);
  }
  
  # パスワード
  &debuglog("# setting password\n");
  $SERIAL_tx_str = "SKSETPWD C $ROUTEB_PWD\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $LOOPSTART = time;
  do {
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_str =~ /OK/ || $LOOPSTART + $EXECTTL < time);
  if($LOOPSTART + $EXECTTL < time) {
    &errorexit("timeout", __LINE__);
  }
  
  
  # デバイススキャン
  &debuglog("# scanning device\n");
  my $rb_ch, $rb_pan, $rb_addr;
  $SERIAL_tx_str = "SKSCAN 2 FFFFFFFF $ROUTEB_SCANTTL 0\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $SERIAL_rx_str = "";
  my $SERIAL_tx_log;
  $LOOPSTART = time;
  # 応答待ち
  do {
    sleep(ROUTEB_SCANWAIT);
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
#    } else {
#      &debuglog(".");
    }
    # たまたま受信タイミングで応答文が途切れる対策（"EVENT 22"を受信するまでループ）
    if($SERIAL_rx_bytes > 0) {
      $SERIAL_rx_log .= $SERIAL_rx_str;
    }
    # ここのタイムアウト時間はデバイス規定の計算方法を元に計算
  } until($SERIAL_rx_log =~ /EVENT 22 / || $LOOPSTART + 0.096 * ( 2 ^ $ROUTEB_SCANTTL + 1) + $EXECTTL < time);
  if($LOOPSTART + 0.096 * ( 2 ^ $ROUTEB_SCANTTL + 1) + $EXECTTL < time) {
    &errorexit("timeout", __LINE__);
  }
  # 応答文を解析
  foreach(split(/\n/,$SERIAL_rx_log)) {
    # チャネル
    if($_ =~ /^  Channel\:([0-9A-F]{2})/) {
      $rb_ch = $1;
      &debuglog("# rb_ch = $rb_ch\n");
    # PanID
    } elsif($_ =~ /^  Pan ID\:([0-9A-F]{4})/) {
      $rb_pan = $1;
      &debuglog("# rb_pan = $rb_pan\n");
    # アドレス
    } elsif($_ =~ /^  Addr\:([0-9A-F]{16})/) {
      $rb_addr = $1;
      &debuglog("# rb_addr = $rb_addr\n");
    } 
  }
  # 解析不可
  if($rb_ch eq "" || $rb_pan eq "" || $rb_addr eq "") {
    &errorexit("cannot get parameter: rb_ch=$rb_ch rb_pan=$rb_pan rb_addr=$rb_addr", __LINE__);
  }

  # アドレス変換(SKLL64)
  &debuglog("# converting rb_addr\n");
  $SERIAL_tx_str = "SKLL64 $rb_addr\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $SERIAL_rx_log = "";
  $LOOPSTART = time;
  do {
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      $SERIAL_rx_log .= $SERIAL_rx_str;
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_log =~ /([0-9A-F:]{39})/ || $LOOPSTART + $EXECTTL < time);
  if($LOOPSTART + $EXECTTL < time) {
    &errorexit("timeout", __LINE__);
  }
  if($SERIAL_rx_log =~ /([0-9A-F:]{39})/) {
    $rb_addr = $1;
    &debuglog("# rb_addr = $rb_addr\n");
  } else {
    &errorexit("cannot get parameter: rb_addr=$rb_addr", __LINE__);
  }

  # チャンネル設定(SKSREG S2)
  &debuglog("# setting channel\n");
  $SERIAL_tx_str = "SKSREG S2 $rb_ch\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $LOOPSTART = time;
  do {
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_str =~ /OK/ || $LOOPSTART + $EXECTTL < time);
  if($LOOPSTART + $EXECTTL < time) {
    die "timeout";
  }

  # Pan ID設定(SKSREG S3)
  &debuglog("# setting PAN ID\n");
  $SERIAL_tx_str = "SKSREG S3 $rb_pan\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $LOOPSTART = time;
  do {
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_str =~ /OK/ || $LOOPSTART + $EXECTTL < time);
  if($LOOPSTART + $EXECTTL < time) {
    &errorexit("timeout", __LINE__);
  }

  # 接続(SKJOIN)
  &debuglog("# connecting\n");
  $SERIAL_tx_str = "SKJOIN $rb_addr\r\n";
  &debuglog("> $SERIAL_tx_str");
  $SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str)
    or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
  $LOOPSTART = time;
  do {
    sleep $ROUTEB_SCANWAIT;
    ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
      or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
    if($SERIAL_rx_bytes > 0) {
      &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
    } else {
#      &debuglog(".");
    }
  } until($SERIAL_rx_str =~ /EVENT 25/ || $LOOPSTART + $EXECTTL < time);
  if($LOOPSTART + $EXECTTL < time) {
    &errorexit("timeout", __LINE__);
  }
  
  # 一時ファイルにアドレスを記録
  &debuglog("# writing rb_addr cached file\n");
  open FILE_RB_ADDR, ">$ROUTEB_TEMPFILE"
    or &errorexit("cannot open file $ROUTEB_TEMPFFILE $!", __LINE__);
  print FILE_RB_ADDR $rb_addr;
  close FILE_RB_ADDR;
  &debuglog("# written file $ROUTEB_TEMPFILE: $rb_addr\n");

}



### データ取得データフレーム作成

# ECHONET Lite データフレーム
#   ECHONET Lite SPECIFICATION
#   第2部 ECHONET Lite通信ミドルウェア仕様
#   第3章 電文構成（フレームフォーマット）

# EHD1: ECHONET Lite電文ヘッダ1 [1byte] (3.2.1.1)
#   0001=ECHONET Lite規格
#   0000=future reserved
$ELFR_EHD1 = pack("B*","00010000");

# EHD2: ECHONET Lite電文ヘッダ2 [1byte] (3.2.1.2)
#   0x81=形式1（規定電文形式）
$ELFR_EHD2 = "\x81";

# TID: Transaction ID [2bytes] (3.2.2)
$ELFR_TID = "\x00\x01";

# SEOJ: 送信元ECHONET Liteオブジェクト指定 [3bytes] (3.2.4)
#   0x05=管理・操作関連機器クラスグループ (表3-1)
#   0xFF=コントローラ (表3-7)
#   インスタンスコード:EOJ識別用コード (3.2.4)
#   応答フレームではSEOJとDEOJが逆になって戻ってくる
$ELFR_SEOJ = "\x05\xFF\x01";

# DEOJ: 相手先ECHONET Liteオブジェクト指定 [3bytes] (3.2.4)
#   0x02=住宅・設備関連機器クラスグループ (表3-1)
#   0x88=低圧スマート電力量メータクラス規定
#        ECHONET SPECIFICATION APPENDIX
#        ECHONET機器オブジェクト詳細規定
#        第3章機器オブジェクト詳細規定
#        3.3.25
#   インスタンスコード:EOJ識別用コード (3.2.4)
$ELFR_DEOJ = "\x02\x88\x01";

# ESV: ECHONET Liteサービス [1byte] (3.2.5)
#   0x62=Get:プロパティ値読み出し要求 (表3-9)
#   応答フレームでは0x72になって戻ってくる
$ELFR_ESV = "\x62";

# OPC: 処理プロパティ数 [1byte] (3.2.6)
#   処理対象プロパティ数
#   EPC1の後で定義する

# EPC1: ECHONET Liteプロパティ1 [1byte] (APPENDIX 3.3.25)
#   ※主なもののみ記載
#   0x80=動作状態
#   0xD3=積算電力計測値・履歴を実使用量に換算する係数
#   0xD7=積算電力量計測値の有効桁数
#   0xE0=積算電力量計測値（正方向）
#   0xE1=積算電力量単位（乗率）
#   0xE3=積算電力量計測値（逆方向・売電）
#   0xE7=瞬時電力計測値（電力実効値）
#   0xE8=瞬時電流計測値（実効電流値）
#   OPCで指定した数だけEPC,PDC,EDTを繰り返し列挙できる
#   九州電力のメーターでは7個以上指定しても6個目までしか返してこなかった
@ELFR_EPC = ("\xD3","\xD7","\xE0","\xE1","\xE7","\xE8");
# EPCの数からOPCを定義する
$ELFR_OPC = &dec2hexpack($#ELFR_EPC + 1, 2);

# PDC1: EDTのバイト数 [1byte] (3.2.8)
#   読み出し要求の場合 0x00
@ELFR_PDC = ("\x00","\x00","\x00","\x00","\x00","\x00","\x00");

# EDT1: プロパティ値データ [PDC] (3.2.9)
#   読み出し要求の場合 空
@ELFR_EDT = ("","","","","","","");


### 読み出し

# フレーム生成
$SERIAL_tx_bin = &ELFR_joint;

# コマンドの生成(SKSENDTO)
$SERIAL_tx_str = "SKSENDTO 1 $rb_addr 0E1A 1 0 " . sprintf("%04X", length($SERIAL_tx_bin)) . " ";

# コマンド送信
&debuglog("# reading\n");
$SERIAL_tx_bytes = $objtty->write($SERIAL_tx_str . $SERIAL_tx_bin . "\r\n")
  or &errorexit("cannot write device $SERIAL_dev $!", __LINE__);
&debuglog("> [$SERIAL_tx_bytes]$SERIAL_tx_str" . unpack("H*", $SERIAL_tx_bin) . "\n");
$SERIAL_rx_log = "";
$LOOPSTART = time;
my $ELFR_rx_chk = uc(unpack("H*",$ELFR_EHD1 . $ELFR_EHD2 . $ELFR_TID . $ELFR_DEOJ . $ELFR_SEOJ . "\x72"));
# 応答(ERXUDP)待ち
do {
  sleep 0.5;
  ($SERIAL_rx_bytes, $SERIAL_rx_str) = $objtty->read($SERIAL_bytes)
    or &errorexit("cannot read device $SERIAL_dev $!", __LINE__);
  &debuglog("< [$SERIAL_rx_bytes]$SERIAL_rx_str\n");
  $SERIAL_rx_log .= $SERIAL_rx_str;
} until($SERIAL_rx_log =~ /ERXUDP .* $ELFR_rx_chk[0-9A-F]+\r\n/ || $SERIAL_rx_log =~ /FAIL ER\d\d/ || $LOOPSTART + $EXECTTL < time);
if($LOOPSTART + $EXECTTL < time) {
  &errorexit("timeout", __LINE__);
}


### 解析

my @get_epc, %get_pdc, %get_edc, $i = 0;
if($SERIAL_rx_log =~ /ERXUDP .* $ELFR_rx_chk([A-F0-9]+)\r\n/) {
  $SERIAL_rx_get = substr($1, 2);
  # プロパティ判別
  while($SERIAL_rx_get ne '') {
    &debuglog("# $SERIAL_rx_get\n");
    # プロパティEPC
    push @get_epc, substr($SERIAL_rx_get, 0, 2);
    # プロパティデータサイズ
    $get_pdc{$get_epc[$i]} = &unpack_unsignedlong(substr($SERIAL_rx_get, 2, 2));
    # プロパティデータ
    $get_edc{$get_epc[$i]} = substr($SERIAL_rx_get, 2 + 2, $get_pdc{$get_epc[$i]} * 2);
    # 残りのデータ
    $SERIAL_rx_get = substr($SERIAL_rx_get, 2 + 2 + $get_pdc{$get_epc[$i]} * 2);
    &debuglog("# $get_epc[$i] [$get_pdc{$get_epc[$i]}] = $get_edc{$get_epc[$i]}\n");
    $i++;
  }
} else {
  # 何もデータが来なかった
  &debuglog("# nodata\n");
}


# 取得データ整形
#   0x80=動作状態
#        unsigned-char(0x30=ON 0x31=OFF 1byte)
if($get_edc{'80'} eq '30') {
  $get_edc{'80'} = '1';
} elsif($get_edc{'80'} eq '31') {
  $get_edc{'80'} = '0';
} else {
  $get_edc{'80'} = '';
}
&debuglog("# 80 = $get_edc{'80'}\n");
#   0xD3=積算電力計測値・履歴を実使用量に換算する係数
#        unsigned-long (0x00000000(000000)-0x000F423F(999999) 4bytes)
if($get_edc{'D3'} =~ /^[0-9A-F]{8}$/) {
  $get_edc{'D3'} = &unpack_unsignedlong($get_edc{'D3'});
} else {
  $get_edc{'D3'} = 1;
}
&debuglog("# D3 = $get_edc{'D3'}\n");
#   0xD7=積算電力量計測値の有効桁数
#        unsigned-char 0x01(1)-0x08(8) 1byte)
$get_edc{'D7'} = &unpack_unsignedlong($get_edc{'D7'});
&debuglog("# D7 = $get_edc{'D7'}\n");
#   0xE1=積算電力量単位（乗率）
#        unsigned-char(0x00(1kWh)-0x04(0.0001kWh) 0x0A(10kWh)-0x0D(10000kWh) 1byte)
$get_edc{'E1'} = &unpack_unsignedlong($get_edc{'E1'});
if($get_edc{'E1'} >= 0x00 && $get_edc{'E1'} <= 0x04) {
  $get_edc{'E1'} = 0.1 ** $get_edc{'E1'};
} elsif($get_edc{'E1'} >= 0x0A && $get_edc{'E1'} <= 0x0D) {
  $get_edc{'E1'} = 10 ** ($get_edc{'E1'} - 0x09);
} else {
  $get_edc{'E1'} = 1;
}
&debuglog("# E1 = $get_edc{'E1'}\n");
#   0xE0=積算電力量計測値（正方向）
#        unsigned-long(0x00000000(0)-0x05F5E0FF(99,999,999) 4bytes)
$get_edc{'E0'} = &unpack_unsignedlong($get_edc{'E0'}) * $get_edc{'E1'};
&debuglog("# E0 = $get_edc{'E0'}\n");
#   0xE3=積算電力量計測値（逆方向・売電）
$get_edc{'E3'} = &unpack_unsignedlong($get_edc{'E3'}) * $get_edc{'E1'};
&debuglog("# E3 = $get_edc{'E3'}\n");
#   0xE7=瞬時電力計測値（電力実効値）
#        1W signed-long(0x80000001(-2,147,483,647)-0x7FFFFFFD 4bytes)
$get_edc{'E7'} = &unpack_signedlong($get_edc{'E7'});
&debuglog("# E7 = $get_edc{'E7'}\n");
#   0xE8=瞬時電流計測値（実効電流値）
#        0.1A R相T相 signed-short(0x8001(-3,276.7)-0x7FFD(3,276.5) 2bytes) x 2
#          0x7FFE=無効
$get_edc{'E8_R'} = &unpack_signedshort(substr($get_edc{'E8'}, 0, 4)) * 0.1;
if(substr($get_edc{'E8'}, 4, 4) eq '7FFE') {
  $get_edc{'E8_T'} = 0;
} else {
  $get_edc{'E8_T'} = &unpack_signedshort(substr($get_edc{'E8'}, 4, 4)) * 0.1;
}
&debuglog("# E8_R = $get_edc{'E8_R'} E8_T = $get_edc{'E8_T'}\n");
&debuglog("\n# finishing\n");

# 出力
print time . "\n";
print "$get_edc{'80'}\n";
print "$get_edc{'D3'}\n";
print "$get_edc{'D7'}\n";
print "$get_edc{'E0'}\n";
print "$get_edc{'E1'}\n";
print "$get_edc{'E3'}\n";
print "$get_edc{'E7'}\n";
print "$get_edc{'E8_R'}\n";
print "$get_edc{'E8_T'}\n";


exit 0;


### サブルーチン

# ECHONET Lite フレーム生成
sub ELFR_joint {
  my $tmp = $ELFR_EHD1 . $ELFR_EHD2 . $ELFR_TID;
  $tmp .= $ELFR_SEOJ . $ELFR_DEOJ . $ELFR_ESV . $ELFR_OPC;
  for($i = 0; $i <= $#ELFR_EPC; $i++) {
    $tmp .= $ELFR_EPC[$i] . $ELFR_PDC[$i] . $ELFR_EDT[$i];
  }
  return $tmp;
}

# 整数を16進数文字列に変換してpack
# ex: &dec2hexpack(5,2) -> 0x05（整数5を2桁の16進数に変換してpack）
sub dec2hexpack {
  return pack("H*",sprintf("%0" . $_[1] . "X", $_[0]));
}

# 16進数文字列をsigned-longとして解釈して10進数に変換
sub unpack_signedlong {
  return unpack("l",pack("L",hex($_[0])));
}

# 16進数文字列をsigned-shortとして解釈して10進数に変換
sub unpack_signedshort {
  return unpack("s",pack("S",hex($_[0])));
}

# 16進数文字列をunsigned-longとして解釈して10進数に変換
sub unpack_unsignedlong {
  return hex($_[0]);
}

# デバッグログ出力・蓄積
sub debuglog {
  $DEBUGLOG .= "$_[0]";
  if($SW_DEBUG == 1) { print "$_[0]"; }
  return;
}

# エラーログを吐いて終了
# ex: &errorexit("error message", __LINE__);
# $DEBUGLOGも併せて吐く。
sub errorexit {
  open ERRORLOG, ">>$ROUTEB_ERRORLOG"
    or die "ERROR: cannot write to log file $ROUTEB_ERRORLOG: $!";
  print ERRORLOG "\n#### ERROR OCCURED " . localtime . ": $_[0] at line $_[1]\n";
  print ERRORLOG $DEBUGLOG;
  close ERRORLOG;
  exit 1;
}

